/* PROLOG Java Version 0.1
Sergey Izwozchikow Copyright 1992, 1996 */

import java.lang.*;
import java.util.*;

import Varname;
import Concret;

class Term extends Object {

// type constants
  static final int Atom = 0;
  static final int Variable = 1;
  static final int Functor = 2;

// reserved Term names
  String invisible = "_";
  String listname = ".";
  String listend = "[]";

// Name of term
  String name;
// type of term
  int type;
// args number
  int reps;
// args array
  Vector args;

  Term()
  {
    name = null;
    type = Term.Atom;
    reps = 0;
    args = null;
  }

  Term(Term source)
  {
    this();
    name = new String(source.name);
    type = source.type;
    reps = source.reps;
    if (type == Term.Functor && reps > 0) {
      args = new Vector(reps);
      for (int i = 0; i < reps; i++)
        args.addElement(new Term((Term) source.args.elementAt(i)));
    }
    else
      args = null;
  }

  int variable(char[] s)
  {
    Character c = new Character(s[0]);
    if (s[0] == '_' || c.isUpperCase(s[0])) return Term.Variable;
    return Term.Atom;
  }

  static final int strchr(char[] src, char c)
  {
    for (int i = 0; i < src.length; i++)
      if (src[i] == c) return i;
    return -1;
  }

  static final int strrchr(char[] src, char c)
  {
    for (int i = src.length - 1; i > 0; i--) 
      if (src[i] == c) return i;
    return -1;
  }

  static final int strchri(char[] src, char c, int indx)
  {
    for (int i = indx; i < src.length; i++)
      if (src[i] == c) return i;
    return -1;
  }

  static final int get(char[] src)
  {
    int brk = strchr(src,(char) '(');
    int spr = strchr(src,(char) ',');
    int dvn = strchr(src,(char) '|');
    int end = strchr(src,(char) ')');
    int lst1 = strchr(src,(char) '[');
    int lst2 = strchr(src,(char) ']');
    int ret = -1;

    if (end == -1) end = strchr( src,(char) '.');
    if (lst1 != -1 && lst2 != -1)
    {
      int cnt = 1;

      for (int l = lst1 + 1; l < src.length; l++)
      {
        if (src[l] == ((char) '[')) cnt++;
        if (src[l] == ((char) ']')) cnt--;
        if (cnt == 0)
        {
          lst2 = l;
          break;
        }
      }
      while (spr != -1 && spr > lst1 && spr < lst2) {
        spr = strchri(src,(char) ',', spr + 1);
      }
      ret = lst2 + 1;
      if( brk != -1 && brk > lst1 && brk < lst2)
        brk = strchri(src,(char) '(', ret);
      if( dvn != -1 && dvn > lst1 && dvn < lst2)
        dvn = strchri(src,(char) '|', ret);
      if( end != -1 && end > lst1 && end < lst2)
        end = strchri(src,(char) ')', ret);
    }
    if (end == -1) end = strrchr( src,(char) ']');
    if (spr == -1) spr = end;
    if (dvn == -1) dvn = end;
    if (brk == -1 || spr < brk || dvn < brk)
      if( brk == -1)
        ret = (spr < dvn) ? spr : dvn;
      else
        ret = (spr < brk) ? spr : dvn;
    else
    {
      int cnt_brk = 1, p;

      for (p = brk + 1; p < src.length; p++)
      {
        if (src[p] == ((char)'(')) cnt_brk++;
        if (src[p] == ((char)')')) cnt_brk--;
        if (cnt_brk == 0)
        {
          ret = p + 1;
          break;
        }
      }
      if (p == src.length)
        ret = -1;
    }
    return ret;
  }

/*
  int args_cnt(char[] src)
  {
    int count = 0;
    int bgn = 0, s = -1;
    String str;

    while (bgn != src.length)
    {
      str = new String(src, bgn, src.length - bgn);
      if ((s = get(str.toCharArray())) != -1)
      {
        s += bgn;
        if( src[s] == ',' || src[s] == ')') count++;
        if( src[s] == ')') break;
      }
      bgn = s + 1;
    }
    return count;
  }
*/

  void read(char[] src)
  {
    int brk, s, dst;

    if( src == null) throw new NullPointerException();
    if( src[0] == ((char)'[') && (strchr( src,(char) ']') != -1))
    {
      String Name = new String(src);
      if (Name.equals(listend))
      {
        name = Name;
        type = Term.Atom;
        reps = 0;
        args = null;
      }
      else
        this.list(src);
    }
    else
      if ((brk = strchr(src,(char) '(')) == -1)
      {
        type = variable(src);
        name = new String(src);
        reps = 0;
        args = null;
      }
      else
      {
        String arg;
        Term targ;
        type = Term.Functor;
        name = new String(src, 0, brk);
        brk = brk + 1;
        args = new Vector();
        while (brk < src.length)
        {
          arg = new String(src, brk, src.length - brk);
          if ((dst = get(arg.toCharArray())) != -1)
          {
            dst += brk;
            arg = new String( src, brk, dst - brk);
            targ = new Term();
            targ.read(arg.toCharArray());
            args.addElement(targ);
            if( src[dst] == ((char)')')) break;
          }
          brk = dst + 1;
        }
        reps = args.size();
      }
  }

  void list(char[] src)
  {
    Term cur = this, a;
    int bgn = strchr( src,(char)'[') + 1, end;
    String lst, s;

    while(true)
    {
      lst = new String(src, bgn, src.length - bgn);
      if ((end = get(lst.toCharArray())) != -1)
      {
        end += bgn;
        cur.name = new String(Term.listname);
        cur.type = Term.Functor;
        cur.reps = 2;
        cur.args = new Vector(2);
        s = new String( src, bgn, end - bgn);
        a = new Term();
        a.read(s.toCharArray());
        cur.args.addElement(a);
        a = new Term();
        cur.args.addElement(a);
        cur = a;
        bgn = end + 1;
      }
      else {
        String er = new String(src);
        throw new RuntimeException("Wrong list in string" + er);
      }
      if( src[end] == ((char)'|'))
      {
        lst = new String(src, bgn, src.length - bgn);
        end = get(lst.toCharArray());
        end += bgn;
        s = new String( src, bgn, end - bgn);
        cur.read(s.toCharArray());
        break;
      }
      if( src[end] == ((char)']'))
      {
        cur.name = new String(listend);
        cur.type = Term.Atom;
        cur.reps = 0;
        cur.args = null;
        break;
      }
    }
  }

  boolean cmp(Term target)
  {
    boolean cmp = true;

    if (target == null) throw new NullPointerException();
    if (this.type == Term.Functor && target.type == Term.Functor)
      if (this.name.equals(target.name) && this.reps == target.reps)
        for( int i = 0; i < this.reps && cmp; i++)
          cmp = ((Term) this.args.elementAt(i)).
                cmp((Term) target.args.elementAt(i));
      else cmp = false;
    else
      if( this.type != Term.Functor && target.type != Term.Functor)
        cmp = this.name.equals(target.name);
      else cmp = false;
    return cmp;
  }

  boolean compare(Term d, Vector concrets, long step)
  {
    boolean comp = true;
    if (d == null) throw new NullPointerException();
    if (this.type == Term.Atom && d.type == Term.Atom)
      comp = this.name.equals(d.name);
    else
      if (this.type == Term.Functor && d.type == Term.Functor)
      {
        if (this.name.equals(d.name) && this.reps == d.reps)
          for (int i = 0; i < this.reps && comp; i++)
            comp = ((Term) this.args.elementAt(i)).
                   compare((Term) d.args.elementAt(i), concrets, step);
        else comp = false;
      }
      else
        if (this.type == Term.Variable && d.type == Term.Variable)
        {
          if (this.name.equals(Term.invisible) ||
              d.name.equals(Term.invisible))
            comp = true;
          else
            if ((comp = this.agree(d, concrets, step)) == true)
              concrets.addElement(new Concret(this.name, d, step));
        }
        else
          if (this.type == Term.Variable || d.type == Term.Variable)
          {
            if (this.name.equals(Term.invisible) ||
                d.name.equals(Term.invisible)) /* invisible variable */
              comp = true;
            else
            {
              if (this.type == Term.Variable)
                comp = this.look(d, concrets, step);
              else
                comp = d.look(this, concrets, step);
            }
          }
          else /* this.type == Term.Functor && d.type == Term.Atom ||
                  this.type == Term.Atom && d.type == Term.Functor */
            comp = false;
    return comp;
  }

  boolean look(Term mean, Vector concrets, long step)
  {
    boolean good = true, find = false;

    for (int i = 0; i < concrets.size() && good; i++)
    {
      if (((Concret) concrets.elementAt(i)).name.equals(this.name))
      {
        if (((Concret) concrets.elementAt(i)).object.type == Term.Variable)
          good = ((Concret) concrets.elementAt(i)).object.
                 look(mean, concrets, step);
        else
          good = ((Concret) concrets.elementAt(i)).object.
                 compare(mean, concrets, step);
        find = true;
      }
    }
    if (find == false)
      concrets.addElement(new Concret(this.name, mean, step));
    return good;
  }

  boolean agree(Term d, Vector concrets, long step)
  {
    boolean conform = true;

    for (int i = 0; i < concrets.size() && conform; i++) {
      Concret c = ((Concret) concrets.elementAt(i));
      if (this.name.equals(c.name))
        if (d.cmp(c.object) == false)
          if (c.object.type ==  Term.Variable)
            conform = c.object.agree(d, concrets, step);
          else
            for (int j = 0; j < concrets.size() && conform; j++) {
              Concret e = ((Concret) concrets.elementAt(j));
              if (d.name.equals(e.name))
                if (this.cmp(e.object) == false)
                  if (e.object.type == Term.Variable)
                    conform = e.object.agree(this, concrets, step);
                  else
                    conform = c.object.compare(e.object, concrets, step);
            }
    }
    return conform;
  }

  boolean like(Term targ)
  {
    boolean similar = true;

    if (this.type == Term.Variable || targ.type == Term.Variable)
      similar = true;
    else
      if (this.type != targ.type) similar = false;
      else
        if (this.name.equals(targ.name) && this.reps == targ.reps)
          similar = true;
        else
          similar = false;
    return similar;
  }

  String generate(long step)
  {
//    String s = new String(step);
//    String rand = new String(valueOf((int)(Math.random() * Integer.MAX_VALUE)));
    return new String("__" + String.valueOf((int)(Math.random() * Integer.MAX_VALUE))
                      + "/"+ String.valueOf(step));
  }

  void diffrent(Vector varnames, long step)
  {
    Varname v = null;
    String n = null;
    int j;

    if (this.type == Term.Functor)
      for (int i = 0; i < this.reps; i++)
        ((Term) args.elementAt(i)).diffrent(varnames, step);
    else
      if (this.type == Term.Variable)
        if (this.name.equals("_") == false) /* visible variable */
        {
          if (varnames.size() == 0)
          {
            n = this.generate(step);
            v = new Varname(this.name, n);
            varnames.addElement(v);
            this.name = new String(n);
          }
          else
          {
            boolean finded = false;
            for (j = 0; j < varnames.size(); j++)
            {
              v = (Varname) varnames.elementAt(j);
              if (v.name.equals(this.name))
              {
                finded = true;
                break;
              }
            }
            if (finded == false)
            {
              n = this.generate(step);
              v = new Varname(this.name, n);
              varnames.addElement(v);
              this.name = new String(n);
            }
            else
              this.name = new String(v.diffname);
          }
        }
  }

  void print()
  {
    System.out.print(this.name);
    if (this.type == Term.Functor)
    {
      int r = this.reps;
      System.out.print("(");
      for (int i = 0; i < r; i++)
      {
        ((Term) args.elementAt(i)).print();
        if (i < r - 1)
          System.out.print(", ");
      }
      System.out.print(")");
    }
  }
}
